# threading.py: run async functions
# (Based on https://github.com/bottlesdevs/Bottles/blob/23d6f190cb680baa55dabc0b321be0ade8a44f1c/bottles/frontend/utils/threading.py)
#
# Copyright 2022 brombinmirko <send@mirko.pm>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, in version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import faulthandler
import logging
import threading
from typing import Any, Callable, Optional, cast

from upscaler.app_profile import PROFILE  # type: ignore

from gi.repository import GLib


class RunAsync(threading.Thread):
    """
    Execute a function asynchronously.

    It takes a function, a callback and a list of arguments as input.
    """

    def __init__(
        self,
        task_func: Callable[..., Any],
        callback: Optional[Callable[[Any, Optional[Exception]], None]],
        *args: Any,
        **kwargs: Any,
    ) -> None:
        if PROFILE == "development":
            faulthandler.enable()

        self.source_id: Optional[int] = None
        assert threading.current_thread() is threading.main_thread()

        super(RunAsync, self).__init__(target=self.__target, args=args, kwargs=kwargs)

        self.task_func = task_func

        self.callback = cast(Callable[..., Any], callback)
        if callback is None:
            self.callback = lambda *args: None

        self.daemon = kwargs.pop("daemon", True)

        self.start()

    def __target(self, *args: Any, **kwargs: Any) -> int:
        result = None
        error = None

        logging.debug(f"Running async job [{self.task_func}].")

        try:
            result = self.task_func(*args, **kwargs)
        except Exception as exception:
            logging.error(
                "Error while running async job: "
                f"{self.task_func}\nException: {exception}"
            )

            error = exception

        self.source_id = GLib.idle_add(self.callback, result, error)
        return self.source_id
